from typing import Type

import graphene

from graphene_extensions.utils import Singleton


class CallableScalarFactory(metaclass=Singleton):
    registered_scalars = {}

    def get(self, scalar: Type[graphene.Scalar]):
        assert issubclass(scalar, graphene.Scalar), f'Callable fields have to of type Scalar, received {scalar}'

        if scalar in self.registered_scalars:
            return self.registered_scalars[scalar]
        return self.register_scalar(scalar)

    def register_scalar(self, scalar: Type[graphene.Scalar]):
        class_name = f'Callable{scalar.__name__}'

        class CallableScalar(scalar):
            class Meta:
                name = class_name

            @staticmethod
            def serialize_callable(value):
                return scalar.serialize(value())

            serialize = serialize_callable

        CallableScalar.__name__ = class_name
        self.registered_scalars[scalar] = CallableScalar
        return self.registered_scalars[scalar]



import json
from collections import OrderedDict
from typing import Tuple, Optional, Union, Callable

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.handlers.wsgi import WSGIRequest
from django.shortcuts import render
from django.views import View
from django.http import HttpResponse, HttpResponseNotAllowed, HttpResponseBadRequest

from graphql.type.schema import GraphQLSchema

from .query import get_query_from_request, QueryExecutor
from .status import STATUS_200_OK, STATUS_400_BAD_REQUEST


class GraphQLView(View):
    allowed_methods = ('GET', 'POST')

    graphiql_version: str = '0.11.10'
    graphiql_template: str = 'graphiql.html'

    schema: Union[GraphQLSchema, Callable] = None

    def dispatch(self, request: WSGIRequest, *args, **kwargs) -> HttpResponse:
        if request.method not in self.allowed_methods:
            return HttpResponseNotAllowed(['GET', 'POST'], 'GraphQL only supports GET and POST requests.')

        try:
            query = self.get_query()
        except ValidationError as e:
            return HttpResponseBadRequest(e.message)

        result, status_code = self.get_query_result(query)

        if self.show_graphiql:
            context = self.get_context_data()
            context.update({'result': result if query else '', 'query': query})
            return render(self.request, self.graphiql_template, context=context,
                          status=status_code if query else STATUS_200_OK)

        return HttpResponse(content=result, status=status_code, content_type='application/json')

    def get_query(self) -> str:
        return get_query_from_request(self.request)

    def execute_query(self, query: str) -> dict:
        return QueryExecutor(self.get_schema()).query(query)

    def get_query_result(self, query: str) -> Tuple[Optional[str], int]:
        result = self.execute_query(query)
        return self.jsonify(result), STATUS_400_BAD_REQUEST if 'errors' in result else STATUS_200_OK

    @classmethod
    def jsonify(cls, data: dict) -> str:
        if settings.DEBUG:
            return json.dumps(data, sort_keys=True, indent=2, separators=(',', ': '))
        return json.dumps(data, separators=(',', ':'))

    def get_schema(self):
        return self.schema(self.request) if callable(self.schema) else self.schema

    @property
    def show_graphiql(self) -> bool:
        return settings.DEBUG and self.request.content_type in ('text/plain', 'text/html')

    def get_context_data(self) -> OrderedDict:
        return OrderedDict({
            'title': 'GraphiQL',
            'graphiql_version': self.graphiql_version,
        })



import json

from graphql import Source, parse, validate, GraphQLSchema
from graphql.error import format_error, GraphQLSyntaxError
from graphql.execution import execute, ExecutionResult

from django.core.exceptions import ValidationError
from django.core.handlers.wsgi import WSGIRequest


class QueryExecutor:
    def __init__(self, schema: GraphQLSchema) -> None:
        self.schema = schema
        assert isinstance(schema, GraphQLSchema), f'schema has to be of type GraphQLSchema, not {self.schema}'

    def query(self, query: str) -> dict:
        assert isinstance(query, str), f'Expected query string, got {query}'

        result = self.execute_query(query)
        if result.invalid:
            return {'errors': [self.format_error(error) for error in result.errors]}
        return {'data': result.data}

    def execute_query(self, query: str) -> ExecutionResult:
        source = Source(query)

        try:
            document_ast = parse(source=source)
        except GraphQLSyntaxError as e:
            return ExecutionResult(errors=[e], invalid=True)

        validation_errors = validate(self.schema, document_ast)
        if validation_errors:
            return ExecutionResult(errors=validation_errors, invalid=True)
        return execute(self.schema, document_ast)

    @classmethod
    def format_error(cls, error: Exception) -> dict:
        if isinstance(error, GraphQLSyntaxError):
            return format_error(error)
        return {'message': str(error)}


def get_query_from_raw_json(data: str) -> str:
    try:
        json_query = json.loads(data)
    except (ValueError, TypeError) as e:
        raise ValueError(e.message)

    if not isinstance(json_query, dict):
        raise ValueError('The received data is not a valid JSON query.')

    if 'query' not in json_query:
        raise ValueError('"query" not in request json')
    return json_query['query']


def get_query_from_request(request: WSGIRequest) -> str:
    if request.method == 'GET':
        return request.GET.get('query', '')

    if not request.content_type:
        raise ValidationError('content-type not specified')

    if request.content_type == 'application/graphql':
        return request.body.decode()
    elif request.content_type == 'application/json':
        try:
            return get_query_from_raw_json(request.body.decode('utf-8'))
        except ValueError as e:
            raise ValidationError(str(e))
    elif request.content_type in ('application/x-www-form-urlencoded', 'multipart/form-data'):
        return request.POST.get('query', '')
    raise ValidationError(f'Unsupported content-type {request.content_type}')



from typing import Type

from graphql.execution.base import ResolveInfo

from graphene.relay import Connection, ConnectionField

from graphene_extensions.optimization import get_queryset
from graphene_extensions.utils.selectors import get_selectors_from_info, strip_relay_selectors

from django.db.models import Manager, Model, QuerySet


class ModelConnectionField(ConnectionField):
    @classmethod
    def connection_resolver(cls, resolver, connection_type, root, info, **kwargs):
        return super().connection_resolver(resolver, connection_type, root, info, connection_info=info, **kwargs)

    @classmethod
    def resolve_connection(cls, connection_type, args, resolved):
        if isinstance(resolved, Manager):
            resolved = resolved.get_queryset()
        else:
            model: Type[Model] = connection_type._meta.node._meta.model
            resolved = cls.get_initial_queryset(model, args['connection_info'])
        return super().resolve_connection(connection_type, args, resolved)

    @property
    def type(self) -> Type[Connection]:
        return self._type._meta.connection

    @classmethod
    def get_initial_queryset(cls, model: Type[Model], info: ResolveInfo) -> QuerySet:
        return get_queryset(model, strip_relay_selectors(get_selectors_from_info(info)))



from typing import Iterable, Type

from django.db.models import Model, QuerySet, Prefetch

from .utils.model_meta import get_model_prefetch, get_model_columns
from .utils.selectors import Selector


def get_queryset(model: Type[Model], selector: Selector) -> QuerySet:
    return model._default_manager.get_queryset() \
        .only(*(get_model_columns(model).intersection(selector.keys()))) \
        .prefetch_related(*get_prefetch(model, selector))


def get_prefetch(model: Type[Model], selector: Selector) -> Iterable[Prefetch]:
    return (
        Prefetch(field_name, queryset=get_queryset(field_model, selector[field_name]))
        for field_name, field_model in get_model_prefetch(model).items() if field_name in selector.keys()
    )



from django.core.management import BaseCommand

from examples.restaurants.fixture import restaurants_fixture


class Command(BaseCommand):
    def handle(self, *args, **options):
        restaurants_fixture()



import inspect
from typing import Type

import graphene
from graphene.types.base import BaseType

from .factories import CallableScalarFactory


def annotate_type(graphene_type: Type[BaseType]):
    assert not isinstance(graphene_type, BaseType), f'Expected type class, got instance: {graphene_type}'
    assert issubclass(graphene_type, BaseType), \
        f'Type annotations have to of subclass graphene.BaseType, received {graphene_type}'

    def wrapper(func):
        if isinstance(func, property):
            return GrapheneProperty(func.fget, func.fget, func.fdel, func.__doc__, graphene_type=graphene_type)
        assert inspect.isfunction(func), f'@annotate_type is only supported for class methods and properties'
        assert issubclass(graphene_type, graphene.Scalar), f'@annotate_type supports only graphene Scalar\'s'
        func._graphene_type = CallableScalarFactory().get(scalar=graphene_type)
        return func

    return wrapper


def graphene_property(graphene_type: Type[BaseType]):
    assert not inspect.isfunction(graphene_type), f'@graphene_property should be initialized with a graphene type'

    assert not isinstance(graphene_type, BaseType), f'Expected type class, got instance: {graphene_type}'
    assert issubclass(graphene_type, BaseType), \
        f'Type annotations have to of subclass graphene.BaseType, received {graphene_type}'

    def wrapper(func):
        return GrapheneProperty(func, graphene_type=graphene_type)

    return wrapper


class GrapheneProperty(property):
    def __init__(self, *args, **kwargs):
        self._graphene_type = kwargs.pop('graphene_type')
        super().__init__(*args, **kwargs)

    def setter(self, fset):
        raise NotImplementedError('graphene_property setter is currently not supported')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-26 23:24
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Owner',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=120)),
            ],
        ),
        migrations.CreateModel(
            name='Pizza',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=120)),
            ],
        ),
        migrations.CreateModel(
            name='Restaurant',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=120)),
                ('owner', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='restaurants.Owner')),
            ],
        ),
        migrations.CreateModel(
            name='Topping',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=120)),
            ],
        ),
        migrations.AddField(
            model_name='pizza',
            name='restaurant',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='pizzas', to='restaurants.Restaurant'),
        ),
        migrations.AddField(
            model_name='pizza',
            name='toppings',
            field=models.ManyToManyField(related_name='pizzas', to='restaurants.Topping'),
        ),
    ]



from django.conf import settings
from django.conf.urls import url, include

urlpatterns = [
    url(r'restaurants', include('examples.restaurants.urls')),
]

if settings.DEBUG:
    import debug_toolbar
    urlpatterns = [
        url(r'^__debug__/', include(debug_toolbar.urls)),
    ] + urlpatterns



import inspect
from typing import Callable, Dict


def get_properties(obj: type, skip=list()) -> Dict[str, property]:
    return {
        name: member for name, member in inspect.getmembers(obj)
        if isinstance(member, property) and name not in skip
    }


def get_methods(obj: object, args_count=1, include_magic_functions: bool = False, skip=list()) -> Dict[str, Callable]:
    return {
        name: member for name, member in inspect.getmembers(obj)
        if inspect.isfunction(member) and name not in skip
        and len(inspect.getfullargspec(member).args) == args_count
        and (name.startswith('__') and name.endswith('__')) == include_magic_functions
    }



from typing import Type
from functools import partial

from django.db.models import Manager, Model, QuerySet

from graphql import ResolveInfo

import graphene

from graphene_extensions.optimization import get_queryset
from graphene_extensions.utils.selectors import strip_relay_selectors, get_selectors_from_info


class ModelListField(graphene.Field):
    def __init__(self, type_, *args, **kwargs):
        super().__init__(graphene.List(type_), *args, **kwargs)

    @property
    def model(self):
        return self.type.of_type._meta.model

    @classmethod
    def list_resolver(cls, resolver, model, root, info, **kwargs):
        resolved = resolver(root, info, **kwargs)
        if not resolved:
            return cls.get_initial_queryset(model, info)
        assert isinstance(resolved, Manager), f'ModelListField expected Manager, got {resolved}'
        return resolved.get_queryset()

    def get_resolver(self, parent_resolver):
        return partial(self.list_resolver, parent_resolver, self.model)

    @classmethod
    def get_initial_queryset(cls, model: Type[Model], info: ResolveInfo) -> QuerySet:
        return get_queryset(model, strip_relay_selectors(get_selectors_from_info(info)))



import collections
from typing import Type, Dict, Any
from inspect import isclass

from django.db import models

import graphene
from graphene import relay
from graphene.types.base import BaseType
from graphene.types.mountedtype import MountedType
from graphene.types.objecttype import ObjectType, ObjectTypeOptions

from graphene_extensions.utils.callables import get_properties, get_methods
from graphene_extensions.utils.model_meta import get_fields
from graphene_extensions.fields import ModelConnectionField, ModelListField

from .registry import TypeRegistry, ModelRegistry


class ModelTypeOptions(ObjectTypeOptions):
    def __init__(self, class_type, model, fields, exclude_fields, connection):
        super().__init__(class_type)
        self.connection: Type[relay.Connection] = connection or self.create_connection()
        self.model: Type[models.Model] = self.resolve_model(model)
        self.fields = self.resolve_fields(model, fields, exclude_fields)
        self.register_model()

    def register_model(self) -> None:
        ModelRegistry().register(self.model, self.class_type)

    def create_connection(self) -> Type[relay.Connection]:
        return relay.Connection.create_type(class_name=f'{self.class_type.__name__}Connection', node=self.class_type)

    def resolve_model(self, model):
        self.validate_model(model)
        return model

    @classmethod
    def validate_model(cls, model):
        assert isclass(model) and issubclass(model, models.Model), \
            f'Meta.model must be a valid Django Model class, received {model}'
        assert not isinstance(model, models.Model), f'Meta.model must be a Django Model class, not instance'

    @classmethod
    def resolve_fields(cls, model: Type[models.Model], fields, exclude_fields) -> Dict[str, BaseType]:
        resolved_fields = collections.OrderedDict()
        model_fields = cls.get_model_fields(model)

        assert not (fields and exclude_fields), f'Using "fields" and "exclude_fields" together is ambiguous'
        if exclude_fields:
            fields = [field for field in model_fields.keys() if field not in exclude_fields]
        if fields == '__all__':
            fields = tuple(model_fields.keys())

        cls.validate_fields(model_fields, fields)
        for field in fields:
            try:
                resolved_fields[field] = cls.get_graphene_type(model_fields[field])
            except AssertionError as e:
                raise AssertionError(f'{field}: {e}')
        return resolved_fields

    @classmethod
    def get_graphene_type(cls, field: models.Field) -> MountedType:
        return TypeRegistry().get(field)

    @classmethod
    def validate_fields(cls, model_fields: dict, fields) -> None:
        message = f'Meta.fields must be either a Tuple/List of model field names or "__all__"'
        if isinstance(fields, (list, tuple)):
            for field in fields:
                assert isinstance(field, str), f'Field name should be of type str, received {field}'
                assert field in model_fields, f'Invalid field "{field}", options are: {tuple(model_fields.keys())}'
        else:
            raise AssertionError(message)

    @classmethod
    def get_model_fields(cls, model: Type[models.Model]) -> Dict[str, Any]:
        return {**get_properties(model),
                **get_methods(model, skip=['clean', 'get_deferred_fields']),
                **get_fields(model)}


class ModelType(ObjectType):
    @classmethod
    def __init_subclass_with_meta__(cls, interfaces=(), possible_types=(), default_resolver=None, _meta=None,
                                    model=None, fields=None, exclude_fields=None, connection=None,
                                    **options):
        if not _meta:
            _meta = ModelTypeOptions(cls, model, fields, exclude_fields, connection)
        cls.validate_meta(_meta)
        super().__init_subclass_with_meta__(interfaces, possible_types, default_resolver, _meta, **options)

    def resolve_id(self, info):  # used to determine ID field when using relay.Node interface
        return self.pk

    @classmethod
    def validate_meta(cls, _meta: ModelTypeOptions) -> None:
        assert isinstance(_meta, ModelTypeOptions), \
            f'class {_meta.__name__} must derive from ModelObjectTypeOptions'

    @classmethod
    def using_relay(cls):
        return any(issubclass(i, graphene.Node) for i in cls._meta.interfaces)

    @classmethod
    def get_field_class(cls, many: bool) -> Type[graphene.Field]:
        if not many:
            return graphene.Field
        return ModelConnectionField if cls.using_relay() else ModelListField



from graphql.execution.base import ResolveInfo
from typing import Dict

from graphql.language.ast import SelectionSet

Selector = Dict[str, Dict]


def get_selectors_from_selection_set(selection_set: SelectionSet) -> Selector:
    return {
        selection.name.value: get_selectors_from_selection_set(selection.selection_set)
        for selection in selection_set.selections
    } if selection_set else None


def get_selectors_from_info(info: ResolveInfo) -> Selector:
    return get_selectors_from_selection_set(info.field_asts[0].selection_set)


def strip_relay_selectors(selector: Selector):
    result_selector = selector.copy()
    for field, selection in result_selector.items():
        if field == 'edges' and isinstance(selection, dict) and 'node' in selection:
            return strip_relay_selectors(selection['node'])
        elif isinstance(selection, dict):
            result_selector[field] = strip_relay_selectors(selection)
    return result_selector



from distutils.core import setup
from setuptools import find_packages

VERSION = '0.1.0b3'

setup(
    name='graphene-extensions',
    version=VERSION,
    description='GraphQL framework for django',
    long_description=open('README.rst').read(),

    author='Karol Gruszczyk',
    author_email='karol.gruszczyk@gmail.com',

    packages=[package for package in find_packages() if 'tests' not in package],

    url='https://github.com/karol-gruszczyk/graphene-extenstions/',
    keywords='graphql framework django relay graphene',

    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='~=3.6',

    install_requires=[
        'graphene>=2.0<3',
        'django>=1.10<2',
    ],
)



from typing import Type, Set, Dict

from django.db.models import Model, Field
from django.db.models.fields.reverse_related import ForeignObjectRel


def get_field_name(field: Field) -> str:
    if isinstance(field, ForeignObjectRel):
        return field.related_name or field.name + ('_set' if field.multiple else '')
    return field.name


def get_related_fields(model: Type[Model]) -> Set[Field]:
    return {field for field in model._meta.get_fields() if field.related_model}


def get_fields(model: Type[Model]) -> Dict[str, Field]:
    fields = {get_field_name(field): field for field in model._meta.get_fields()}
    return {**fields, 'pk': fields[model._meta.pk.name]}


def is_selectable(field: Field) -> bool:
    return field.many_to_one or field.one_to_one and not field.auto_created


def is_prefetchable(field: Field) -> bool:
    return field.many_to_many or field.one_to_one or field.one_to_many or is_selectable(field)


def get_model_select(model: Type[Model]) -> Dict[str, Type[Model]]:
    return {get_field_name(field): field.related_model
            for field in get_related_fields(model) if is_selectable(field)}


def get_model_prefetch(model: Type[Model]) -> Dict[str, Type[Model]]:
    return {get_field_name(field): field.related_model
            for field in get_related_fields(model) if is_prefetchable(field)}


def get_model_columns(model: Type[Model]) -> Set[str]:
    return {field.attname for field in model._meta._forward_fields_map.values() if not field.many_to_many}



from django.db import models

import graphene

from graphene_extensions.decorators import annotate_type, graphene_property


class Owner(models.Model):
    name = models.CharField(max_length=120)

    def __str__(self):
        return self.name


class Restaurant(models.Model):
    owner = models.OneToOneField(Owner)
    name = models.CharField(max_length=120)

    def __str__(self):
        return self.name


class Topping(models.Model):
    name = models.CharField(max_length=120)

    @graphene_property(graphene.String)
    def str(self):
        return str(self)

    def __str__(self):
        return f'{self.name}'


class Pizza(models.Model):
    restaurant = models.ForeignKey(Restaurant, related_name='pizzas')
    name = models.CharField(max_length=120)
    toppings = models.ManyToManyField(Topping, related_name='pizzas')

    @annotate_type(graphene.String)
    def str(self):
        return str(self)

    def __str__(self):
        return f'{self.name}'



import graphene
from graphene_extensions import ModelListField

from graphene_extensions import ModelType

from .models import Owner, Restaurant, Pizza, Topping


class OwnerType(ModelType):
    class Meta:
        model = Owner
        fields = '__all__'


class RestaurantType(ModelType):
    class Meta:
        model = Restaurant
        fields = '__all__'


class PizzaType(ModelType):
    class Meta:
        model = Pizza
        fields = '__all__'


class ToppingType(ModelType):
    class Meta:
        model = Topping
        fields = '__all__'


class Query(graphene.ObjectType):
    owners = ModelListField(OwnerType)
    restaurants = ModelListField(RestaurantType)
    pizzas = ModelListField(PizzaType)
    toppings = ModelListField(ToppingType)


schema = graphene.Schema(query=Query)



from django.db import transaction

from .models import Owner, Restaurant, Pizza, Topping


@transaction.atomic()
def restaurants_fixture():
    if any(model.objects.exists() for model in (Owner, Restaurant, Pizza, Topping)):
        return print('Some restaurant objects already exist, skipping...')

    luigi = Owner.objects.create(name='Luigi')
    john = Owner.objects.create(name='John')

    cheese = Topping.objects.create(name='cheese')
    salami = Topping.objects.create(name='salami')
    basil = Topping.objects.create(name='basil')
    tomato = Topping.objects.create(name='tomato')
    ham = Topping.objects.create(name='ham')
    ananas = Topping.objects.create(name='ananas')

    italian_pizzeria = Restaurant.objects.create(name='Italian pizzeria', owner=luigi)
    new_york_pizza = Restaurant.objects.create(name='New York pizza', owner=john)

    margarita = Pizza.objects.create(name='margarita', restaurant=italian_pizzeria)
    pepperoni = Pizza.objects.create(name='pepperoni', restaurant=italian_pizzeria)
    hawaii = Pizza.objects.create(name='hawaii', restaurant=new_york_pizza)

    margarita.toppings.set((cheese, tomato, basil))
    pepperoni.toppings.set((cheese, tomato, basil, salami))
    hawaii.toppings.set((ananas, tomato, ham, cheese))

    print('Restaurant fixture loaded!')



import inspect
from typing import Type, Dict, Callable, Any

import graphene
from django.contrib.postgres import fields
from django.db import models
from django.db.models.fields.reverse_related import ForeignObjectRel
from graphene.types.base import BaseType
from graphene.types.mountedtype import MountedType

from graphene_extensions.utils.singleton import Singleton


class ModelRegistry(metaclass=Singleton):
    registry: Dict[Type[models.Model], Type[graphene.ObjectType]] = {}

    def register(self, model: Type[models.Model], model_type: Type[graphene.ObjectType]) -> None:
        assert issubclass(model, models.Model)
        assert issubclass(model_type, graphene.ObjectType)
        self.registry[model] = model_type

    def get(self, model: Type[models.Model]) -> Type[graphene.ObjectType]:
        if model not in self.registry:
            raise RuntimeError(f'{model} has no registered ModelObjectType')
        return self.registry[model]

    def clear(self) -> None:
        self.registry = {}


class TypeRegistry(metaclass=Singleton):
    registry: Dict[Type, BaseType] = {}

    foreign_fields = {models.OneToOneField, models.OneToOneRel, models.ForeignKey}
    related_fields = {models.ManyToManyField, models.ManyToManyRel, models.ManyToOneRel} | foreign_fields

    def register(self, _type: Type, graphene_type: Type[BaseType]) -> None:
        assert issubclass(graphene_type, BaseType)
        self.registry[_type] = graphene_type

    def get(self, field: Any) -> MountedType:
        if isinstance(field, (models.Field, ForeignObjectRel)):
            return self.get_model_type(field)
        if isinstance(field, property):
            assert hasattr(field, '_graphene_type'), f'decorate property with @annotate_type or use @graphene_property'
            return graphene.Field(type=field._graphene_type)
        if inspect.isfunction(field):
            assert hasattr(field, '_graphene_type'), f'decorate method with @annotate_type to determine graphene type'
            return graphene.Field(type=field._graphene_type)
        if inspect.isclass(field) and issubclass(field, models.Field):
            raise AssertionError(f'Expected field instance, got class: {field}')
        raise NotImplementedError(f'field conversion for {field} is not supported')

    def get_model_type(self, field: models.Field) -> MountedType:
        field_type = type(field)
        if field_type in self.related_fields:
            return self.get_related_model_type(field)
        if field_type not in self.registry:
            raise NotImplementedError(f'{field_type} field conversion is not implemented')
        return graphene.Field(type=self.registry[field_type])

    def clear(self) -> None:
        self.registry = {}

    @classmethod
    def get_related_model_type(cls, field: models.Field) -> graphene.Dynamic:
        model = field.related_model
        return graphene.Dynamic(cls.get_field_resolver(model, many=type(field) not in cls.foreign_fields))

    @classmethod
    def get_field_resolver(cls, model: Type[models.Model], many: bool) -> Callable:
        def lazy_type():
            model_type = ModelRegistry().get(model)
            return model_type.get_field_class(many)(model_type)

        return lazy_type


def register(_type: Type, graphene_type: Type[BaseType]) -> None:
    TypeRegistry().register(_type, graphene_type)


register(models.AutoField, graphene.ID)
register(models.IntegerField, graphene.Int)
register(models.PositiveIntegerField, graphene.Int)
register(models.SmallIntegerField, graphene.Int)
register(models.PositiveSmallIntegerField, graphene.Int)
register(models.BigIntegerField, graphene.Int)
register(models.FloatField, graphene.Float)

register(models.CharField, graphene.String)
register(models.TextField, graphene.String)
register(models.EmailField, graphene.String)
register(models.FileField, graphene.String)

register(models.DateField, graphene.String)
register(models.DateTimeField, graphene.String)

register(fields.JSONField, graphene.JSONString)
